const vec3 AP1_RGB_Y = vec3(0.2722287168, 0.6740817658, 0.0536895174);


const mat3 AP1_YCC  = mat3(
    AP1_RGB_Y.r                               ,   AP1_RGB_Y.g                               ,   AP1_RGB_Y.b,
    -0.5 * (AP1_RGB_Y.r / (1 - AP1_RGB_Y.b))  ,   -0.5 * (AP1_RGB_Y.g / (1 - AP1_RGB_Y.b))  ,   0.5,
    0.5                                       ,   -0.5 * (AP1_RGB_Y.g / (AP1_RGB_Y.r))      ,   -0.5 * (AP1_RGB_Y.b / (1 - AP1_RGB_Y.r))
);

const mat3 YCC_AP1  = mat3(
    1.0,  0.0                                                     ,  2 - 2 * AP1_RGB_Y.r,
    1.0,  -(AP1_RGB_Y.b / AP1_RGB_Y.g) * (2 - 2 * AP1_RGB_Y.b)    ,  -(AP1_RGB_Y.r / AP1_RGB_Y.g) * (2 - 2 * AP1_RGB_Y.r),
    1.0,  2 - 2 * AP1_RGB_Y.b                                     ,  0.0
);

//const mat3 AP1_YCC = mat3(0.25, 0.5, -0.25, 0.5, 0.0, 0.5, 0.25, -0.5, -0.25);
//const mat3 YCC_AP1 = mat3(1.0, 1.0, 1.0, 1.0, 0.0, -1.0, -1.0, 1.0, -1.0);


struct SH
{
    vec4 shY;
    vec2 CoCg;
};

// Switch to enable or disable the *look* of spherical harmonics lighting.
// Does not affect the performance, just for A/B image comparison.
#define ENABLE_SH 1

vec3 project_SH_irradiance(SH sh, vec3 N)
{
#if ENABLE_SH
    float d = dot(sh.shY.xyz, N);
    float Y = 2.0 * (1.023326 * d + 0.886226 * sh.shY.w);
    Y = max(Y, 0.0);

    sh.CoCg *= Y * 0.282095 / (sh.shY.w + 1e-6);

    float   T       = Y - sh.CoCg.y * 0.5;
    float   G       = sh.CoCg.y + T;
    float   B       = T - sh.CoCg.x * 0.5;
    float   R       = B + sh.CoCg.x;

    return max(vec3(R, G, B), vec3(0.0));
#else
    return sh.shY.xyz;
#endif
}

SH irradiance_to_SH(vec3 color, vec3 dir)
{
    SH result;

    float   Co      = color.r - color.b;
    float   t       = color.b + Co * 0.5;
    float   Cg      = color.g - t;
    float   Y       = max(t + Cg * 0.5, 0.0);

    result.CoCg = vec2(Co, Cg);

    float   L00     = 0.282095;
    float   L1_1    = 0.488603 * dir.y;
    float   L10     = 0.488603 * dir.z;
    float   L11     = 0.488603 * dir.x;

    result.shY = vec4 (L11, L1_1, L10, L00) * Y;

    return result;
}

float getLumaSH(SH sh) {
    return sh.shY.w / 0.282095;
}

vec3 SH_to_irradiance(SH sh)
{
    float   Y       = sh.shY.w / 0.282095;

    float   T       = Y - sh.CoCg.y * 0.5;
    float   G       = sh.CoCg.y + T;
    float   B       = T - sh.CoCg.x * 0.5;
    float   R       = B + sh.CoCg.x;

    return max(vec3(R, G, B), vec3(0.0));
}

SH init_SH()
{
    SH result;
    result.shY = vec4(0);
    result.CoCg = vec2(0);
    return result;
}

void accumulate_SH(inout SH accum, SH b, float scale)
{
    accum.shY += b.shY * scale;
    accum.CoCg += b.CoCg * scale;
}


void scaleSH(inout SH data, float mult) {
    data.shY *= mult;
    data.CoCg *= mult;
}
void divideSH(inout SH data, float denom) {
    data.shY /= denom;
    data.CoCg /= denom;
}

SH mix_SH(SH a, SH b, float s)
{
    SH result;
    result.shY = mix(a.shY, b.shY, vec4(s));
    result.CoCg = mix(a.CoCg, b.CoCg, vec2(s));
    return result;
}

SH load_SH(sampler2D img_shY, sampler2D img_CoCg, ivec2 p)
{
    SH result;
    result.shY = texelFetch(img_shY, p, 0);
    result.CoCg = texelFetch(img_CoCg, p, 0).xy;
    return result;
}